
/* This is an SDL implementation of an ICO to BMP translator

   (Information taken from "Icons in Win32",
	http://www.microsoft.com/win32dev/ui/icons.htm)
 */

#include <stdio.h>

#include "SDL.h"
#include "SDL_endian.h"
#include "seeico.h"

static void PrintBMP(Win32BMP *bmp)
{
	int i, j, k;
	Uint8 *pixels;

	/* Convert raw data to little endian */
	bmp->biSize = SDL_SwapLE32(bmp->biSize);
	bmp->biWidth = SDL_SwapLE32(bmp->biWidth);
	bmp->biHeight = SDL_SwapLE32(bmp->biHeight);
	bmp->biPlanes = SDL_SwapLE16(bmp->biPlanes);
	bmp->biBitCount = SDL_SwapLE16(bmp->biBitCount);
	bmp->biCompression = SDL_SwapLE32(bmp->biCompression);
	bmp->biSizeImage = SDL_SwapLE32(bmp->biSizeImage);
	bmp->biXPelsPerMeter = SDL_SwapLE32(bmp->biXPelsPerMeter);
	bmp->biYPelsPerMeter = SDL_SwapLE32(bmp->biYPelsPerMeter);
	bmp->biClrUsed = SDL_SwapLE32(bmp->biClrUsed);
	bmp->biClrImportant = SDL_SwapLE32(bmp->biClrImportant);

	/* Print values */
	printf("\t\tbiSize = %d\n", bmp->biSize);
	printf("\t\tbiWidth = %d\n", bmp->biWidth);
	printf("\t\tbiHeight = %d\n", bmp->biHeight);
	printf("\t\tbiPlanes = %d\n", bmp->biPlanes);
	printf("\t\tbiBitCount = %d\n", bmp->biBitCount);
	printf("\t\tbiCompression = %d\n", bmp->biCompression);
	printf("\t\tbiSizeImage = %d\n", bmp->biSizeImage);
	printf("\t\tbiXPelsPerMeter = %d\n", bmp->biXPelsPerMeter);
	printf("\t\tbiYPelsPerMeter = %d\n", bmp->biYPelsPerMeter);
	printf("\t\tbiClrUsed = %d\n", bmp->biClrUsed);
	printf("\t\tbiClrImportant = %d\n", bmp->biClrImportant);

	pixels = (Uint8 *)bmp+bmp->biSize+(1<<bmp->biBitCount)*4;
	printf("** Data\n");
	for ( i=0; i<bmp->biHeight/2; ++i ) {
		for ( j=0; j<bmp->biWidth*bmp->biBitCount/8; ++j )
			printf("%2.2X", *pixels++);
		printf("\n");
	}
	printf("** Mask\n");
	for ( i=0; i<bmp->biHeight/2; ++i ) {
		for ( j=0; j<bmp->biWidth/8; ++j ) {
			int ch = *pixels++;
			for ( k=0; k<8; ++k ) {
				printf("%2.2X", ch&0x80 ? 0xFF : 0x00);
				ch <<= 1;
			}
		}
		printf("\n");
	}
	printf("Final position: %d\n", (int)pixels-(int)bmp);
}

int main(int argc, char *argv[])
{
	ICONDIRENTRY ico_entry;
	FILE *ico;
	int numicons;
	long pos, i;
	Win32BMP *bmp;

	/* Initialize SDL */
	if ( SDL_Init(0) < 0 ) {
		fprintf(stderr, "Couldn't initialize SDL: %s\n",SDL_GetError());
		exit(-1);
	}
	atexit(SDL_Quit);

	/* Check command line */
	if ( argc < 2 ) {
		fprintf(stderr, "Usage: %s file.ico\n", argv[0]);
		exit(1);
	}

	/* Open the ICO file, and read the header */
	ico = fopen(argv[1], "rb");
	if ( ico == NULL ) {
		fprintf(stderr, "Couldn't read %s\n", argv[1]);
		exit(2);
	}
	SDL_ReadLE16(ico);
	switch ((i=SDL_ReadLE16(ico))) {
		case 1:
			printf("Collection of icons\n");
			break;
		case 2:
			printf("Collection of cursors\n");
			break;
		default:
			printf("Unknown file format: %d\n", i);
			exit(2);
	}
	numicons = SDL_ReadLE16(ico);
	printf("Number of images: %d\n", numicons);

	/* Read each icon */
	for ( i=0; i<numicons; ++i ) {
		fread(&ico_entry.bWidth, 1, 1, ico);
		fread(&ico_entry.bHeight, 1, 1, ico);
		fread(&ico_entry.bColorCount, 1, 1, ico);
		fread(&ico_entry.bReserved, 1, 1, ico);
		ico_entry.wPlanes = SDL_ReadLE16(ico);
		ico_entry.wBitCount = SDL_ReadLE16(ico);
		ico_entry.dwBytesInRes = SDL_ReadLE32(ico);
		ico_entry.dwImageOffset = SDL_ReadLE32(ico);
		printf("\timage %d: %dx%dx%d -- %d bytes\n", i, ico_entry.bWidth, ico_entry.bHeight, ico_entry.wBitCount ? 1<<ico_entry.wBitCount : ico_entry.bColorCount, ico_entry.dwBytesInRes);
		pos = ftell(ico);
		fseek(ico, ico_entry.dwImageOffset, SEEK_SET);
		bmp = (Win32BMP *)malloc(ico_entry.dwBytesInRes);
		if ( bmp != NULL ) {
			if ( ! fread(bmp, ico_entry.dwBytesInRes, 1, ico) ) {
				fprintf(stderr,
					"WARNING: Couldn't read image\n");
			} else {
				PrintBMP(bmp);
			}
			free(bmp);
		}
		fseek(ico, pos, SEEK_SET);
	}
	exit(0);
}
