/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997, 1998  Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    5635-34 Springhouse Dr.
    Pleasanton, CA 94588 (USA)
    slouken@devolution.com
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_blit_N.c,v 1.4 1998/09/15 16:33:13 slouken Exp $";
#endif

#include <stdio.h>

#include "SDL_types.h"
#include "SDL_video.h"
#include "SDL_blit.h"

/* Functions to blit from N-bit surfaces to other surfaces */

#ifdef USE_ASMBLIT

/* Heheheh, we coerce Hermes into using SDL blit information */
#define HermesInterface	SDL_BlitInfo

#include "Head_MMX.h"
#include "Head_x86.h"

/* Function to check the CPU flags */
#ifdef NO_MMXBLIT
#define MMX_OK()	0
#else
#define MMX_OK()	((CheckMMX() & 0x800000) == 0x800000)
#endif

#else

/* Special optimized blit for RGB 8-8-8 --> RGB 3-3-2 */
#define RGB888_RGB332(dst, src) { \
	dst = (((src)&0x00E00000)>>16)| \
	      (((src)&0x0000E000)>>11)| \
	      (((src)&0x000000C0)>>6); \
}
static void Blit_RGB888_index8(SDL_BlitInfo *info)
{
	int c;
	int width, height;
	Uint32 *src;
	Uint8 *map, *dst;
	int srcskip, dstskip;

	/* Set up some basic variables */
	width = info->d_width;
	height = info->d_height;
	src = (Uint32 *)info->s_pixels;
	srcskip = info->s_skip/4;
	dst = info->d_pixels;
	dstskip = info->d_skip;
	map = info->table;

	if ( map == NULL ) {
		while ( height-- ) {
			for ( c=width/4; c; --c ) {
				/* Pack RGB into 8bit pixel */
				RGB888_RGB332(*dst++, *src);
				++src;
				RGB888_RGB332(*dst++, *src);
				++src;
				RGB888_RGB332(*dst++, *src);
				++src;
				RGB888_RGB332(*dst++, *src);
				++src;
			}
			switch ( width % 4 ) {
				case 3:
					RGB888_RGB332(*dst++, *src);
					++src;
				case 2:
					RGB888_RGB332(*dst++, *src);
					++src;
				case 1:
					RGB888_RGB332(*dst++, *src);
					++src;
			}
			src += srcskip;
			dst += dstskip;
		}
	} else {
		int pixel;

		while ( height-- ) {
			for ( c=width/4; c; --c ) {
				/* Pack RGB into 8bit pixel */
				RGB888_RGB332(pixel, *src);
				*dst++ = map[pixel];
				++src;
				RGB888_RGB332(pixel, *src);
				*dst++ = map[pixel];
				++src;
				RGB888_RGB332(pixel, *src);
				*dst++ = map[pixel];
				++src;
				RGB888_RGB332(pixel, *src);
				*dst++ = map[pixel];
				++src;
			}
			switch ( width % 4 ) {
				case 3:
					RGB888_RGB332(pixel, *src);
					*dst++ = map[pixel];
					++src;
				case 2:
					RGB888_RGB332(pixel, *src);
					*dst++ = map[pixel];
					++src;
				case 1:
					RGB888_RGB332(pixel, *src);
					*dst++ = map[pixel];
					++src;
			}
			src += srcskip;
			dst += dstskip;
		}
	}
}
/* Special optimized blit for RGB 8-8-8 --> RGB 5-5-5 */
#define RGB888_RGB555(dst, src) { \
	*(Uint16 *)(dst) = (((*src)&0x00F80000)>>9)| \
	                   (((*src)&0x0000F800)>>6)| \
	                   (((*src)&0x000000F8)>>3); \
}
#define RGB888_RGB555_TWO(dst, src) { \
	*(Uint32 *)(dst) = (((((src[1])&0x00F80000)>>9)| \
	                     (((src[1])&0x0000F800)>>6)| \
	                     (((src[1])&0x000000F8)>>3))<<16)| \
	                     (((src[0])&0x00F80000)>>9)| \
	                     (((src[0])&0x0000F800)>>6)| \
	                     (((src[0])&0x000000F8)>>3); \
}
static void Blit_RGB888_RGB555(SDL_BlitInfo *info)
{
	int c;
	int width, height;
	Uint32 *src;
	Uint16 *dst;
	int srcskip, dstskip;

	/* Set up some basic variables */
	width = info->d_width;
	height = info->d_height;
	src = (Uint32 *)info->s_pixels;
	srcskip = info->s_skip/4;
	dst = (Uint16 *)info->d_pixels;
	dstskip = info->d_skip/2;

	/* Memory align at 4-byte boundary, if necessary */
	if ( (int)dst & 0x03 ) {
		/* Don't do anything if width is 0 */
		if ( width == 0 ) {
			return;
		}
		--width;

		while ( height-- ) {
			/* Perform copy alignment */
			RGB888_RGB555(dst, src);
			++src;
			++dst;

			/* Copy in 4 pixel chunks */
			for ( c=width/4; c; --c ) {
				RGB888_RGB555_TWO(dst, src);
				src += 2;
				dst += 2;
				RGB888_RGB555_TWO(dst, src);
				src += 2;
				dst += 2;
			}
			/* Get any leftovers */
			switch (width % 4) {
				case 3:
					RGB888_RGB555(dst, src);
					++src;
					++dst;
				case 2:
					RGB888_RGB555_TWO(dst, src);
					src += 2;
					dst += 2;
					break;
				case 1:
					RGB888_RGB555(dst, src);
					++src;
					++dst;
					break;
			}
			src += srcskip;
			dst += dstskip;
		}
	} else { 
		while ( height-- ) {
			/* Copy in 4 pixel chunks */
			for ( c=width/4; c; --c ) {
				RGB888_RGB555_TWO(dst, src);
				src += 2;
				dst += 2;
				RGB888_RGB555_TWO(dst, src);
				src += 2;
				dst += 2;
			}
			/* Get any leftovers */
			switch (width % 4) {
				case 3:
					RGB888_RGB555(dst, src);
					++src;
					++dst;
				case 2:
					RGB888_RGB555_TWO(dst, src);
					src += 2;
					dst += 2;
					break;
				case 1:
					RGB888_RGB555(dst, src);
					++src;
					++dst;
					break;
			}
			src += srcskip;
			dst += dstskip;
		}
	}
}
/* Special optimized blit for RGB 8-8-8 --> RGB 5-6-5 */
#define RGB888_RGB565(dst, src) { \
	*(Uint16 *)(dst) = (((*src)&0x00F80000)>>8)| \
	                   (((*src)&0x0000FC00)>>5)| \
	                   (((*src)&0x000000F8)>>3); \
}
#define RGB888_RGB565_TWO(dst, src) { \
	*(Uint32 *)(dst) = (((((src[1])&0x00F80000)>>8)| \
	                     (((src[1])&0x0000FC00)>>5)| \
	                     (((src[1])&0x000000F8)>>3))<<16)| \
	                     (((src[0])&0x00F80000)>>8)| \
	                     (((src[0])&0x0000FC00)>>5)| \
	                     (((src[0])&0x000000F8)>>3); \
}
static void Blit_RGB888_RGB565(SDL_BlitInfo *info)
{
	int c;
	int width, height;
	Uint32 *src;
	Uint16 *dst;
	int srcskip, dstskip;

	/* Set up some basic variables */
	width = info->d_width;
	height = info->d_height;
	src = (Uint32 *)info->s_pixels;
	srcskip = info->s_skip/4;
	dst = (Uint16 *)info->d_pixels;
	dstskip = info->d_skip/2;

	/* Memory align at 4-byte boundary, if necessary */
	if ( (int)dst & 0x03 ) {
		/* Don't do anything if width is 0 */
		if ( width == 0 ) {
			return;
		}
		--width;

		while ( height-- ) {
			/* Perform copy alignment */
			RGB888_RGB565(dst, src);
			++src;
			++dst;

			/* Copy in 4 pixel chunks */
			for ( c=width/4; c; --c ) {
				RGB888_RGB565_TWO(dst, src);
				src += 2;
				dst += 2;
				RGB888_RGB565_TWO(dst, src);
				src += 2;
				dst += 2;
			}
			/* Get any leftovers */
			switch (width % 4) {
				case 3:
					RGB888_RGB565(dst, src);
					++src;
					++dst;
				case 2:
					RGB888_RGB565_TWO(dst, src);
					src += 2;
					dst += 2;
					break;
				case 1:
					RGB888_RGB565(dst, src);
					++src;
					++dst;
					break;
			}
			src += srcskip;
			dst += dstskip;
		}
	} else { 
		while ( height-- ) {
			/* Copy in 4 pixel chunks */
			for ( c=width/4; c; --c ) {
				RGB888_RGB565_TWO(dst, src);
				src += 2;
				dst += 2;
				RGB888_RGB565_TWO(dst, src);
				src += 2;
				dst += 2;
			}
			/* Get any leftovers */
			switch (width % 4) {
				case 3:
					RGB888_RGB565(dst, src);
					++src;
					++dst;
				case 2:
					RGB888_RGB565_TWO(dst, src);
					src += 2;
					dst += 2;
					break;
				case 1:
					RGB888_RGB565(dst, src);
					++src;
					++dst;
					break;
			}
			src += srcskip;
			dst += dstskip;
		}
	}
}
#endif /* USE_ASMBLITTER */

/* Special optimized blit for RGB 8-8-8 --> RGB 3-3-2 */
#ifndef RGB888_RGB332
#define RGB888_RGB332(dst, src) { \
	dst = (((src)&0x00E00000)>>16)| \
	      (((src)&0x0000E000)>>11)| \
	      (((src)&0x000000C0)>>6); \
}
#endif
static void Blit_RGB888_index8_map(SDL_BlitInfo *info)
{
	int c, pixel;
	int width, height;
	Uint32 *src;
	Uint8 *map, *dst;
	int srcskip, dstskip;

	/* Set up some basic variables */
	width = info->d_width;
	height = info->d_height;
	src = (Uint32 *)info->s_pixels;
	srcskip = info->s_skip/4;
	dst = info->d_pixels;
	dstskip = info->d_skip;
	map = info->table;

	while ( height-- ) {
		for ( c=width/4; c; --c ) {
			/* Pack RGB into 8bit pixel */
			RGB888_RGB332(pixel, *src);
			*dst++ = map[pixel];
			++src;
			RGB888_RGB332(pixel, *src);
			*dst++ = map[pixel];
			++src;
			RGB888_RGB332(pixel, *src);
			*dst++ = map[pixel];
			++src;
			RGB888_RGB332(pixel, *src);
			*dst++ = map[pixel];
			++src;
		}
		switch ( width % 4 ) {
			case 3:
				RGB888_RGB332(pixel, *src);
				*dst++ = map[pixel];
				++src;
			case 2:
				RGB888_RGB332(pixel, *src);
				*dst++ = map[pixel];
				++src;
			case 1:
				RGB888_RGB332(pixel, *src);
				*dst++ = map[pixel];
				++src;
		}
		src += srcskip;
		dst += dstskip;
	}
}
static void SDL_BlitNto1(SDL_BlitInfo *info)
{
	int c;
	int width, height;
	Uint8 *src, *map, *dst;
	int srcskip, dstskip;
	Uint8 srcbpp;
	Uint32 pixel;
	Uint8  sR, sG, sB;
	SDL_PixelFormat *srcfmt;

	/* Set up some basic variables */
	width = info->d_width;
	height = info->d_height;
	src = info->s_pixels;
	srcskip = info->s_skip;
	dst = info->d_pixels;
	dstskip = info->d_skip;
	map = info->table;
	srcfmt = info->src;
	srcbpp = srcfmt->BytesPerPixel;

	if ( map == NULL ) {
		while ( height-- ) {
			for ( c=width; c; --c ) {
				DISEMBLE_RGB(src, srcbpp, srcfmt, pixel,
								sR, sG, sB);
				if ( 1 ) {
				  	/* Pack RGB into 8bit pixel */
				  	*dst = ((sR>>5)<<(3+2))|
						((sG>>5)<<(2)) |
						((sB>>6)<<(0)) ;
				}
				dst++;
				src += srcbpp;
			}
			src += srcskip;
			dst += dstskip;
		}
	} else {
		while ( height-- ) {
			for ( c=width; c; --c ) {
				DISEMBLE_RGB(src, srcbpp, srcfmt, pixel,
								sR, sG, sB);
				if ( 1 ) {
				  	/* Pack RGB into 8bit pixel */
				  	*dst = map[((sR>>5)<<(3+2))|
						   ((sG>>5)<<(2))  |
						   ((sB>>6)<<(0))  ];
				}
				dst++;
				src += srcbpp;
			}
			src += srcskip;
			dst += dstskip;
		}
	}
}
static void SDL_BlitNtoN(SDL_BlitInfo *info)
{
	int c;
	int width, height;
	Uint8 *src, *dst;
	Uint8 srcbpp;
	Uint8 dstbpp;
	int srcskip, dstskip;
	Uint32 pixel;
	Uint8  sR, sG, sB;
	SDL_PixelFormat *srcfmt;
	SDL_PixelFormat *dstfmt;

	/* Set up some basic variables */
	width = info->d_width;
	height = info->d_height;
	src = info->s_pixels;
	srcskip = info->s_skip;
	dst = info->d_pixels;
	dstskip = info->d_skip;
	srcfmt = info->src;
	srcbpp = srcfmt->BytesPerPixel;
	dstfmt = info->dst;
	dstbpp = dstfmt->BytesPerPixel;

	while ( height-- ) {
		for ( c=width; c; --c ) {
			DISEMBLE_RGB(src, srcbpp, srcfmt, pixel, sR, sG, sB);
			if ( 1 ) {
				  ASSEMBLE_RGB(dst, dstbpp, dstfmt, sR, sG, sB);
			}
			dst += dstbpp;
			src += srcbpp;
		}
		src += srcskip;
		dst += dstskip;
	}
}

static SDL_loblit complex_blit[] = {
	NULL, SDL_BlitKey, SDL_BlitAlpha, SDL_BlitAlphaKey
};

SDL_loblit SDL_CalculateBlit2(SDL_Surface *surface, int complex)
{
	if ( complex ) {
		return(complex_blit[complex]);
	} else {
		switch (surface->dstfmt->BitsPerPixel/8) {
			case 0:
				return(NULL);
			case 1:
				return(SDL_BlitNto1);
			default:
				return(SDL_BlitNtoN);
		}
	}
}
SDL_loblit SDL_CalculateBlit3(SDL_Surface *surface, int complex)
{
	if ( complex ) {
		return(complex_blit[complex]);
	} else {
		switch (surface->dstfmt->BitsPerPixel/8) {
			case 0:
				return(NULL);
			case 1:
				return(SDL_BlitNto1);
			default:
				return(SDL_BlitNtoN);
		}
	}
}
SDL_loblit SDL_CalculateBlit4(SDL_Surface *surface, int complex)
{
	SDL_loblit blitfun;

	if ( complex ) {
		return(complex_blit[complex]);
	} else {
		SDL_PixelFormat *srcfmt;
		SDL_PixelFormat *dstfmt;

		srcfmt = surface->format;
		dstfmt = surface->dstfmt;
		if ( surface->dstfmt->BitsPerPixel < 8 ) {
			return(NULL);
		}
		switch (surface->dstfmt->BytesPerPixel) {
			case 1:
				if ( (srcfmt->Rmask == 0x00FF0000) &&
				     (srcfmt->Gmask == 0x0000FF00) &&
				     (srcfmt->Bmask == 0x000000FF) ) {
					if ( surface->palmap ) {
						blitfun=Blit_RGB888_index8_map;
					} else {
#ifdef USE_ASMBLIT
						surface->swaccel_data->aux_data
							= Convert32_8RGB332;
						blitfun = ConvertX86;
#else
						blitfun = Blit_RGB888_index8;
#endif
					}
				} else {
					blitfun = SDL_BlitNto1;
				}
				break;
			case 2:
				if ( (srcfmt->Rmask == 0x00FF0000) &&
				     (srcfmt->Gmask == 0x0000FF00) &&
				     (srcfmt->Bmask == 0x000000FF) &&
				     (dstfmt->Rmask == 0x0000F800) &&
				     (dstfmt->Gmask == 0x000007E0) &&
				     (dstfmt->Bmask == 0x0000001F) ) {
#ifdef USE_ASMBLIT
					if ( MMX_OK() ) {
						surface->swaccel_data->aux_data
							= ConvertMMX32_16RGB565;
						blitfun = ConvertMMX;
					} else {
						surface->swaccel_data->aux_data
							= Convert32_16RGB565;
						blitfun = ConvertX86;
					}
#else
					blitfun = Blit_RGB888_RGB565;
#endif
				} else
				if ( (srcfmt->Rmask == 0x00FF0000) &&
				     (srcfmt->Gmask == 0x0000FF00) &&
				     (srcfmt->Bmask == 0x000000FF) &&
				     (dstfmt->Rmask == 0x0000001F) &&
				     (dstfmt->Gmask == 0x000007E0) &&
				     (dstfmt->Bmask == 0x0000F800) ) {
#ifdef USE_ASMBLIT
					surface->swaccel_data->aux_data
							= Convert32_16BGR565;
					blitfun = ConvertX86;
#else
					blitfun = SDL_BlitNtoN;
#endif
				} else
				if ( (srcfmt->Rmask == 0x00FF0000) &&
				     (srcfmt->Gmask == 0x0000FF00) &&
				     (srcfmt->Bmask == 0x000000FF) &&
				     (dstfmt->Rmask == 0x00007C00) &&
				     (dstfmt->Gmask == 0x000003E0) &&
				     (dstfmt->Bmask == 0x0000001F) ) {
#ifdef USE_ASMBLIT
					if ( MMX_OK() ) {
						surface->swaccel_data->aux_data
							= ConvertMMX32_16RGB555;
						blitfun = ConvertMMX;
					} else {
						surface->swaccel_data->aux_data
							= Convert32_16RGB555;
						blitfun = ConvertX86;
					}
#else
					blitfun = Blit_RGB888_RGB555;
#endif
				} else
				if ( (srcfmt->Rmask == 0x00FF0000) &&
				     (srcfmt->Gmask == 0x0000FF00) &&
				     (srcfmt->Bmask == 0x000000FF) &&
				     (dstfmt->Rmask == 0x0000001F) &&
				     (dstfmt->Gmask == 0x000003E0) &&
				     (dstfmt->Bmask == 0x00007C00) ) {
#ifdef USE_ASMBLIT
					surface->swaccel_data->aux_data
							= Convert32_16BGR555;
					blitfun = ConvertX86;
#else
					blitfun = SDL_BlitNtoN;
#endif
				} else {
					blitfun = SDL_BlitNtoN;
				}
				break;
			case 3:
				if ( (srcfmt->Rmask == 0x00FF0000) &&
				     (srcfmt->Gmask == 0x0000FF00) &&
				     (srcfmt->Bmask == 0x000000FF) &&
				     (dstfmt->Rmask == 0x00FF0000) &&
				     (dstfmt->Gmask == 0x0000FF00) &&
				     (dstfmt->Bmask == 0x000000FF) ) {
#ifdef USE_ASMBLIT
						surface->swaccel_data->aux_data
							= Convert32_24RGB888;
						blitfun = ConvertX86;
#else
						blitfun = SDL_BlitNtoN;
#endif
				} else
				if ( (srcfmt->Rmask == 0x00FF0000) &&
				     (srcfmt->Gmask == 0x0000FF00) &&
				     (srcfmt->Bmask == 0x000000FF) &&
				     (dstfmt->Rmask == 0x000000FF) &&
				     (dstfmt->Gmask == 0x0000FF00) &&
				     (dstfmt->Bmask == 0x00FF0000) ) {
#ifdef USE_ASMBLIT
						surface->swaccel_data->aux_data
							= Convert32_24BGR888;
						blitfun = ConvertX86;
#else
						blitfun = SDL_BlitNtoN;
#endif
				} else {
					blitfun = SDL_BlitNtoN;
				}
				break;
			default:
				blitfun = SDL_BlitNtoN;
		}
	}

#ifdef ASM_DEBUG
#ifdef USE_ASMBLIT
	if ( blitfun == ConvertMMX )
		fprintf(stderr, "Using mmx blit\n");
	else
	if ( blitfun == ConvertX86 )
		fprintf(stderr, "Using asm blit\n");
	else
#endif
	if ( blitfun == SDL_BlitNtoN )
		fprintf(stderr, "Using C blit\n");
	else
		fprintf(stderr, "Using optimized C blit\n");
#endif /* ASM_DEBUG */

	return(blitfun);
}
