/*
    DRAW: An example program for the SDL, maclib, and screenlib libraries
    Copyright (C) 1997  Sam Lantinga

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    5635-34 Springhouse Dr.
    Pleasanton, CA 94588 (USA)
    slouken@devolution.com
*/

class Sprite {

public:
	Sprite(FrameBuf *Screen, SDL_Surface *Image,
				Sint16 X, Sint16 Y, Sint16 Xvel, Sint16 Yvel) {
		screen = Screen;
		image = Image;
		SetPosition(X, Y);
		SetVelocity(Xvel, Yvel);
	}

	virtual ~Sprite() { }

	virtual void Hide(void) {
		screen->Clear(x, y, image->w, image->h);
	}
	virtual void Show(void) {
		screen->QueueBlit(x, y, image);
	}
	virtual void Move(void) {
		x += xvel;
		if ( x < 0 ) {
			x = -x;
			xvel = -xvel;
		} else
		if ( (x+image->w) >= screen->Width() ) {
			x -= ((x+image->w+1) - screen->Width());
			xvel = -xvel;
		}
		y += yvel;
		if ( y < 0 ) {
			y = -y;
			yvel = -yvel;
		} else
		if ( (y+image->h) >= screen->Height() ) {
			y -= ((y+image->h+1) - screen->Height());
			yvel = -yvel;
		}
	}

	virtual void SetPosition(Sint16 X, Sint16 Y) {
		x = X-image->w/2;
		y = Y-image->h/2;
	}
	virtual void GetPosition(Sint16 &X, Sint16 &Y) {
		X = x+image->w/2;
		Y = y+image->h/2;
	}
	virtual void SetVelocity(Sint16 Xvel, Sint16 Yvel) {
		xvel = Xvel;
		yvel = Yvel;
	}
	virtual void GetVelocity(Sint16 &Xvel, Sint16 &Yvel) {
		Xvel = xvel;
		Yvel = yvel;
	}

private:
	FrameBuf *screen;
	SDL_Surface *image;
	Sint16 x, y;
	Sint16 xvel, yvel;
};
